@@ -21,6 +21,7 @@ use Action::*;
 pub struct Registers {
     ep: u32,
     a: u32,
+    memory: u32,
     sp: u32,
     rp: u32,
     s0: u32,
@@ -35,7 +36,7 @@ pub struct Registers {
 #[derive(Default)]
 struct AllRegisters {
     public: Registers,
-    memory: u64,
+    m0: u64,
     opcode: u32,
     stack0: u32,
     stack1: u32,
@@ -121,6 +122,10 @@ impl<T: Target> VM<T> {
             free_cells: memory_cells,
             halt_addr: 0,
         };
+        // Set memory size register.
+        vm.registers_mut().memory = u32::try_from(
+            cell_bytes(i64::from(memory_cells))
+        ).expect("Address out of range");
         // Allocate the data stack.
         let sp = vm.allocate(data_cells).1;
         vm.registers_mut().s0 = sp;
@@ -131,7 +136,7 @@ impl<T: Target> VM<T> {
         vm.registers_mut().rp = rp;
         // Allocate a word to hold a HALT instruction.
         vm.halt_addr = vm.allocate(1).0;
-        vm.store(vm.halt_addr, 0x5F);
+        vm.store(vm.halt_addr, 0x55);
         vm
     }
 
@@ -215,7 +220,7 @@ impl<T: Target> VM<T> {
     pub unsafe fn run(mut self, ep: u32) -> Self {
         assert!(Self::is_aligned(ep));
         self.registers_mut().ep = ep;
-        self.state.memory = self.memory.as_mut_ptr() as u64;
+        self.state.m0 = self.memory.as_mut_ptr() as u64;
         *self.jit.global_mut(code::Global(0)) = Word {mp: (&mut self.state as *mut AllRegisters).cast()};
         let (jit, trap) = self.jit.execute(&State::Root).expect("Execute failed");
         assert_eq!(trap, Trap::Halt);
@@ -246,7 +251,7 @@ const BEP: Variable = Variable::Register(REGISTERS[6]);
 const BA: Variable = Variable::Register(REGISTERS[7]);
 const BSP: Variable = Variable::Register(REGISTERS[8]);
 const BRP: Variable = Variable::Register(REGISTERS[9]);
-const MEMORY: Variable = Variable::Register(REGISTERS[10]);
+const M0: Variable = Variable::Register(REGISTERS[10]);
 const OPCODE: Variable = Variable::Register(REGISTERS[11]);
 const STACK0: Variable = Variable::Slot(Slot(0));
 const STACK1: Variable = Variable::Slot(Slot(1));
@@ -366,7 +371,7 @@ impl Builder {
      * `TEMP` is corrupted.
      */
     fn native_address(&mut self, dest: impl IntoVariable, addr: impl IntoVariable) {
-        self.binary64(Add, dest, MEMORY, addr);
+        self.binary64(Add, dest, M0, addr);
     }
 
     /**
@@ -513,7 +518,7 @@ impl code::Machine for Machine {
     fn num_globals(&self) -> usize { 1 }
 
     fn marshal(&self, state: Self::State) -> Marshal {
-        let mut live_values = vec![Global(0).into(), BEP, BSP, BRP, MEMORY];
+        let mut live_values = vec![Global(0).into(), BEP, BSP, BRP, M0];
         #[allow(clippy::match_same_arms)]
         live_values.extend(match state {
             State::Root => vec![BA],
@@ -542,7 +547,7 @@ impl code::Machine for Machine {
             b.load_register(BA, public_register!(a));
             b.load_register(BSP, public_register!(sp));
             b.load_register(BRP, public_register!(rp));
-            b.load_register64(MEMORY, private_register!(memory));
+            b.load_register64(M0, private_register!(m0));
             b.load_register(OPCODE, private_register!(opcode));
             b.load_register(STACK0, private_register!(stack0));
             b.load_register(STACK1, private_register!(stack1));
@@ -561,7 +566,7 @@ impl code::Machine for Machine {
             b.store_register(BA, public_register!(a));
             b.store_register(BSP, public_register!(sp));
             b.store_register(BRP, public_register!(rp));
-            b.store_register64(MEMORY, private_register!(memory));
+            b.store_register64(M0, private_register!(m0));
             b.store_register(OPCODE, private_register!(opcode));
             b.store_register(STACK0, private_register!(stack0));
             b.store_register(STACK1, private_register!(stack1));
@@ -621,7 +626,7 @@ impl code::Machine for Machine {
             State::Lshift => Switch::if_(
                 OPCODE, // `Ult(STACK0, CELL_BITS)`
                 build(|b| {
-                    b.binary(Lsl, R2, STACK0, STACK1);
+                    b.binary(Lsl, R2, STACK1, STACK0);
                     b.store(R2, BSP);
                 }, Ok(State::Root)),
                 build(|b| {
@@ -632,7 +637,7 @@ impl code::Machine for Machine {
             State::Rshift => Switch::if_(
                 OPCODE, // `Ult(STACK0, CELL_BITS)`
                 build(|b| {
-                    b.binary(Lsr, R2, STACK0, STACK1);
+                    b.binary(Lsr, R2, STACK1, STACK0);
                     b.store(R2, BSP);
                 }, Ok(State::Root)),
                 build(|b| {
@@ -706,7 +711,7 @@ impl code::Machine for Machine {
                 build(|b| {
                     // Discard the loop index and limit.
                     b.const_binary(Add, BRP, BRP, cell_bytes(2));
-                }, Ok(State::Root)),
+                }, Ok(State::Next)),
                 build(|_| {}, Ok(State::Branchi)),
             ),
             State::Ploopi => Switch::if_(
@@ -940,7 +945,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(Add, R2, R2, R4);
+                        b.binary(Add, R2, R4, R2);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -948,7 +953,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(Sub, R2, R2, R4);
+                        b.binary(Sub, R2, R4, R2);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -956,7 +961,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(Sub, R2, R4, R2);
+                        b.binary(Sub, R2, R2, R4);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -992,7 +997,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(Mul, R2, R2, R4);
+                        b.binary(Mul, R2, R4, R2);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -1053,7 +1058,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(Max, R2, R2, R4);
+                        b.binary(Max, R2, R4, R2);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -1061,7 +1066,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(Min, R2, R2, R4);
+                        b.binary(Min, R2, R4, R2);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -1076,7 +1081,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(And, R2, R2, R4);
+                        b.binary(And, R2, R4, R2);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -1084,7 +1089,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(Or, R2, R2, R4);
+                        b.binary(Or, R2, R4, R2);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -1092,7 +1097,7 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(R2, BSP);
                         b.load(R4, BSP);
-                        b.binary(Xor, R2, R2, R4);
+                        b.binary(Xor, R2, R4, R2);
                         b.store(R2, BSP);
                     }, Ok(State::Root)),
 
@@ -1100,14 +1105,14 @@ impl code::Machine for Machine {
                     build(|b| {
                         b.pop(STACK0, BSP);
                         b.load(STACK1, BSP);
-                        b.const_binary(Ult, OPCODE, STACK1, CELL_BITS);
+                        b.const_binary(Ult, OPCODE, STACK0, CELL_BITS);
                     }, Ok(State::Lshift)),
 
                     // RSHIFT
                     build(|b| {
                         b.pop(STACK0, BSP);
                         b.load(STACK1, BSP);
-                        b.const_binary(Ult, OPCODE, STACK1, CELL_BITS);
+                        b.const_binary(Ult, OPCODE, STACK0, CELL_BITS);
                     }, Ok(State::Rshift)),
 
                     // 1LSHIFT
@@ -1161,86 +1166,27 @@ impl code::Machine for Machine {
                         b.store(R3, R2);
                     }, Ok(State::Root)),
 
-                    // BSP@
+                    // SP@
                     build(|b| {
                         b.move_(R1, BSP);
                         b.push(R1, BSP);
                     }, Ok(State::Root)),
 
-                    // BSP!
+                    // SP!
                     build(|b| {
                         b.load(BSP, BSP);
                     }, Ok(State::Root)),
 
-                    // BRP@
+                    // RP@
                     build(|b| {
                         b.push(BRP, BSP);
                     }, Ok(State::Root)),
 
-                    // BRP!
+                    // RP!
                     build(|b| {
                         b.pop(BRP, BSP);
                     }, Ok(State::Root)),
 
-                    // EP@
-                    build(|b| {
-                        b.push(BEP, BSP);
-                    }, Ok(State::Root)),
-
-                    // BS0@
-                    build(|b| {
-                        b.load_register(R1, public_register!(s0));
-                        b.push(R1, BSP);
-                    }, Ok(State::Root)),
-
-                    // BS0!
-                    build(|b| {
-                        b.pop(R1, BSP);
-                        b.store_register(R1, public_register!(s0));
-                    }, Ok(State::Root)),
-
-                    // BR0@
-                    build(|b| {
-                        b.load_register(R1, public_register!(r0));
-                        b.push(R1, BSP);
-                    }, Ok(State::Root)),
-
-                    // BR0!
-                    build(|b| {
-                        b.pop(R1, BSP);
-                        b.store_register(R1, public_register!(r0));
-                    }, Ok(State::Root)),
-
-                    // 'THROW@
-                    build(|b| {
-                        b.load_register(R1, public_register!(throw));
-                        b.push(R1, BSP);
-                    }, Ok(State::Root)),
-
-                    // 'THROW!
-                    build(|b| {
-                        b.pop(R1, BSP);
-                        b.store_register(R1, public_register!(throw));
-                    }, Ok(State::Root)),
-
-                    // MEMORY@
-                    build(|b| {
-                        b.load_register(R1, private_register!(memory));
-                        b.push(R1, BSP);
-                    }, Ok(State::Root)),
-
-                    // 'BAD@
-                    build(|b| {
-                        b.load_register(R1, public_register!(bad));
-                        b.push(R1, BSP);
-                    }, Ok(State::Root)),
-
-                    // -ADDRESS@
-                    build(|b| {
-                        b.load_register(R1, public_register!(not_address));
-                        b.push(R1, BSP);
-                    }, Ok(State::Root)),
-
                     // BRANCH
                     build(|_| {}, Ok(State::Branch)),
 
@@ -1386,8 +1332,82 @@ impl code::Machine for Machine {
 
                     // HALT
                     build(|_| {}, Err(Trap::Halt)),
+
+                    // EP@
+                    build(|b| {
+                        b.push(BEP, BSP);
+                    }, Ok(State::Root)),
+
+                    // LIB
+                    build(|_| {
+                        // TODO.
+                    }, Err(Trap::NotImplemented)),
+
+                    // UNDEFINED
+                    build(|_| {
+                        // TODO.
+                    }, Err(Trap::NotImplemented)),
+
+                    // LINK
+                    build(|_| {
+                        // TODO.
+                    }, Err(Trap::NotImplemented)),
+
+                    // S0@
+                    build(|b| {
+                        b.load_register(R1, public_register!(s0));
+                        b.push(R1, BSP);
+                    }, Ok(State::Root)),
+
+                    // S0!
+                    build(|b| {
+                        b.pop(R1, BSP);
+                        b.store_register(R1, public_register!(s0));
+                    }, Ok(State::Root)),
+
+                    // R0@
+                    build(|b| {
+                        b.load_register(R1, public_register!(r0));
+                        b.push(R1, BSP);
+                    }, Ok(State::Root)),
+
+                    // R0!
+                    build(|b| {
+                        b.pop(R1, BSP);
+                        b.store_register(R1, public_register!(r0));
+                    }, Ok(State::Root)),
+
+                    // 'THROW@
+                    build(|b| {
+                        b.load_register(R1, public_register!(throw));
+                        b.push(R1, BSP);
+                    }, Ok(State::Root)),
+
+                    // 'THROW!
+                    build(|b| {
+                        b.pop(R1, BSP);
+                        b.store_register(R1, public_register!(throw));
+                    }, Ok(State::Root)),
+
+                    // MEMORY@
+                    build(|b| {
+                        b.load_register(R1, public_register!(memory));
+                        b.push(R1, BSP);
+                    }, Ok(State::Root)),
+
+                    // 'BAD@
+                    build(|b| {
+                        b.load_register(R1, public_register!(bad));
+                        b.push(R1, BSP);
+                    }, Ok(State::Root)),
+
+                    // -ADDRESS@
+                    build(|b| {
+                        b.load_register(R1, public_register!(not_address));
+                        b.push(R1, BSP);
+                    }, Ok(State::Root)),
                 ]),
-                build(|_| {}, Err(Trap::Halt)),
+                build(|_| {}, Err(Trap::NotImplemented)),
             ),
         }
     }
@@ -1445,10 +1465,10 @@ pub mod tests {
 
         // Beetle object code:
         vec![
-            0x00001504, 0x0000024F, 0x00002108, 0x0000084D,
-            0x00001501, 0x0000034F, 0x001A2202, 0xFFFFF853,
-            0x0000034D, 0x22062204, 0xFFFFF553, 0xFFFFF453,
-            0x00000054,
+            0x00001504, 0x00000245, 0x00002108, 0x00000843,
+            0x00001501, 0x00000345, 0x001A2202, 0xFFFFF849,
+            0x00000343, 0x22062204, 0xFFFFF549, 0xFFFFF449,
+            0x0000004A,
         ]
     }
 
