@@ -336,12 +336,19 @@ impl<B: Buffer> Lowerer<B> {
 
     /** Select how to assemble a shift `BinaryOp` such as `Shl`. */
     fn shift_binary(&mut self, op: ShiftOp, prec: Precision, dest: impl Into<Register>, src1: impl Into<Value>, src2: impl Into<Value>) {
-        let dest = dest.into();
-        let src1 = src1.into();
+        let mut dest = dest.into();
+        let mut src1 = src1.into();
         let src2 = src2.into();
-        let save_rc = src2 != Value::Register(RC);
+        let save_rc = dest == RC || src2 != Value::Register(RC);
         if save_rc {
-            self.move_(TEMP, RC);
+            if dest == RC {
+                dest = TEMP;
+            } else {
+                self.move_(TEMP, RC);
+            }
+            if src1 == Value::Register(RC) {
+                src1 = Value::Register(TEMP);
+            }
         }
         let src2 = self.src_to_register(src2, RC);
         self.move_(RC, src2);
@@ -755,4 +762,45 @@ pub mod tests {
     fn constants() {
         assert_eq!(CONSTANTS[ZERO_ADDRESS / size_of::<Word>()], Word {u: 0});
     }
+
+    #[test]
+    fn shift_binary() {
+        let pool = Pool::new(0);
+        let mut lo = Lowerer::<Vec<u8>>::new(pool);
+        let start = lo.here().target().unwrap();
+        for (dest, src1, src2) in [
+            (RC, RA, RC),
+            (RC, RA, RD),
+            (RD, RA, RC),
+            (RD, RA, RD),
+            (RA, RC, RD),
+        ] {
+            lo.shift_binary(Shl, P32, dest, src1, src2);
+        }
+        disassemble(&lo.a, start, vec![
+            "mov r12,rax",
+            "shl r12d,cl",
+            "mov rcx,r12",
+
+            "mov rcx,rdx",
+            "mov r12,rax",
+            "shl r12d,cl",
+            "mov rcx,r12",
+
+            "mov rdx,rax",
+            "shl edx,cl",
+
+            "mov r12,rcx",
+            "mov rcx,rdx",
+            "mov rdx,rax",
+            "shl edx,cl",
+            "mov rcx,r12",
+
+            "mov r12,rcx",
+            "mov rcx,rdx",
+            "mov rax,r12",
+            "shl eax,cl",
+            "mov rcx,r12",
+        ]).unwrap();
+    }
 }
